import { dirname, resolve, join, basename } from "path";
import { mkdir, copyFile, symlink, access as _access } from "fs";
import { constants } from "os";

export const _exists = async (path: string) => new Promise<boolean>((_resolve, reject) => {
  _access(path, error => {
    if (error) {
      return _resolve(false)
    }
    return _resolve(true)
  })
})

export const _failIfNotExists = async (path: string) => {
  const exists = await _exists(path)
  if (!exists) {
    throw `${path} does not exist`
  }
}
/**
 * @param path to make recursively
 */
export const _make = async (path: string) => new Promise<void>((_resolve, reject) => {
  if (0 === path.length || '.' === path || '/' === path) {
    return _resolve()
  }
  _make(dirname(path)).then(() => _exists(path)).then(exists => {
    if (exists) {
      return _resolve()
    }
    mkdir(path, error => {
      if (error) {
        return reject(error)
      }
      return _resolve()
    })
  })
})
/**
 * NOTE that destination is overwritten if it already exists
 */
export const _copy = async (source: string, destination: string) => new Promise<void>((_resolve, reject) => {
  _make(dirname(destination)).then(() => {
    copyFile(source, destination, error => {
      if (error) {
        return reject(error)
      }
      return _resolve()
    })
  })
})
/**
 * NOTE that it won't fail in case target exist
 */
export const _sl = async (source: string, target: string, type: 'dir' | 'file', log?: (message:string) => void) => new Promise<void>((_resolve, reject) => {
  // not sure what it does when path is not absolute path
  symlink(resolve(source), resolve(target), type, error => {
    if (error) {
      if ('EEXIST' === error.code) {
        if (log) {
          log(`${target} already exists, EEXIST:${constants.errno.EEXIST}, error:${error.errno}`)
        }
        return _resolve()
      }
      return reject(error)
    }
    return _resolve()
  })
})
/**
 * search for a basename(@path) starting from dirname(@path) up to process.cwd()
 * @param path name to search for, e.g. package.json
 * @returns absolute path to existing basename(@path) or null in case it do not exists all the way up to process.cwd()
 */
export const _search = async (path:string):Promise<string|null> => {
  const directory = resolve(dirname(path))
  // is it inside process.cwd()
  if (!directory.startsWith(process.cwd())) {
    return null
  }
  const exists = await _exists(path)
  if (exists) {
    // resolve as path has to be absolute
    return resolve(path)
  }
  return await _search(join(dirname(directory),basename(path)))
}